package de.devloop.mavor.servlet;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.util.Objects;
import java.util.UUID;
import java.util.stream.Collectors;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import de.devloop.mavor.Artifact;
import de.devloop.mavor.AuthenticatedServlet;
import jakarta.servlet.RequestDispatcher;
import jakarta.servlet.ServletException;
import jakarta.servlet.annotation.WebServlet;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;

@WebServlet("/download/jars")
public class DownloadJars extends AuthenticatedServlet {

    private static final String PARAMETER_REPOSITORY = "repository";
    private static final String PARAMETER_GROUP_ID = "groupId";
    private static final String PARAMETER_ARTIFACT_ID = "artifactId";
    private static final String PARAMETER_VERSION = "version";
    private static final String PARAMETER_POM = "pom";
    private static final String PARAMETER_TYPE = "type";

    // TODO: set as property
    private static final String MAVEN_CMD = "/usr/bin/mvn";
    private static final String TEMP_DIR = "/home/damage/Temp/mavor";

    private final UUID randomUUID = UUID.randomUUID();
    private final File tempDir = new File(TEMP_DIR);
    private final File tempDirJars = new File(tempDir, randomUUID.toString());

    @Override
    protected void doAuthenticatedPost(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        doAuthenticatedGet(req, resp);
    }

    @Override
    protected void doAuthenticatedGet(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        // create UUID subdirectory
        // should only be one level, so not mkdirs
        tempDirJars.mkdir();
        
        // create pom.xml
        File pomXmlFile = setupPomFile(req);

        // run maven command
        ExecutionResult executionResult;
        try {
            executionResult = executeMaven();
        } catch (InterruptedException e) {
            throw new ServletException("Running maven failed", e);
        } finally {
            pomXmlFile.delete();
        }

        // and off we go
        req.setAttribute("stdout", executionResult.stdout);
        req.setAttribute("stderr", executionResult.stderr);
        req.setAttribute("exitcode", executionResult.exitCode);

        String zipFilename = zipDependencies();
        req.setAttribute("zipFilename", zipFilename);

        RequestDispatcher view = req.getRequestDispatcher("/download.jsp");
        view.forward(req, resp);
    }

    private Artifact getArtifact(HttpServletRequest req) {
        String repository = Objects.requireNonNull(req.getParameter(PARAMETER_REPOSITORY));
        String groupId = Objects.requireNonNull(req.getParameter(PARAMETER_GROUP_ID));
        String artifactId = Objects.requireNonNull(req.getParameter(PARAMETER_ARTIFACT_ID));
        String version = Objects.requireNonNull(req.getParameter(PARAMETER_VERSION));

        Artifact artifact = new Artifact();
        artifact.setGroupId(groupId);
        artifact.setArtifactId(artifactId);
        artifact.setVersion(version);
        artifact.setRepository(repository);

        return artifact;
    }

    private ExecutionResult executeMaven() throws IOException, InterruptedException {
        Process mvnProcess = Runtime.getRuntime().exec(MAVEN_CMD + " dependency:copy-dependencies -DoutputDirectory=" + tempDirJars.getCanonicalPath(), null, tempDirJars.getCanonicalFile());
        String stdout = new String(mvnProcess.getInputStream().readAllBytes(), StandardCharsets.UTF_8);
        String stderr = new String(mvnProcess.getErrorStream().readAllBytes(), StandardCharsets.UTF_8);

        ExecutionResult executionResult = new ExecutionResult();
        
        executionResult.exitCode = mvnProcess.waitFor();
        executionResult.stdout = stdout;
        executionResult.stderr = stderr;
        return executionResult;
    }

    private File setupPomFile(HttpServletRequest req) throws ServletException, IOException {
        File pomXmlFile;
        try {
            String type = Objects.requireNonNull(req.getParameter(PARAMETER_TYPE));
            if (type.equals("artifact")) {
                Artifact artifact = getArtifact(req);
                pomXmlFile = generatePomXmlByArtifact(artifact);
            } else if (type.equals("pom")) {
                String pomXml = Objects.requireNonNull(req.getParameter(PARAMETER_POM));
                pomXmlFile = writePomXml(pomXml);
            } else {
                throw new ServletException("Unknown download type: " + type);
            }
        } catch (NullPointerException npe) {
            throw new ServletException("Missing required parameter", npe);
        }

        return pomXmlFile;
    }

    private File generatePomXmlByArtifact(Artifact artifact) throws IOException {
        String pomXml = getResourceFileAsString("pomstub.xml");
        pomXml = pomXml.replace("{repository-id}", "foo");
        pomXml = pomXml.replace("{repository-url}", artifact.getRepository());
        pomXml = pomXml.replace("{group-id}", artifact.getGroupId());
        pomXml = pomXml.replace("{artifact-id}", artifact.getArtifactId());
        pomXml = pomXml.replace("{version}", artifact.getVersion());
        return writePomXml(pomXml);
    }

    private File writePomXml(String pomXml) throws IOException {
        File pomXmlFile = new File(tempDirJars, "pom.xml");
        FileWriter fileWriter = new FileWriter(pomXmlFile);
        fileWriter.write(pomXml);
        fileWriter.close();

        return pomXmlFile;
    }

    private String zipDependencies() throws IOException {
        File zipFile = new File(tempDir, "mavor_" + UUID.randomUUID() + ".zip");
        ZipOutputStream zipStream = new ZipOutputStream(new FileOutputStream(zipFile));

        // get all jar files, make sure it is a jar and in correct directory
        File[] jarFiles = tempDirJars.listFiles(new FilenameFilter() {

            @Override
            public boolean accept(File dir, String name) {
                try {
                    return dir.getCanonicalPath().equals(tempDirJars.getCanonicalPath()) && name.endsWith(".jar");
                } catch (IOException e) {
                    return false;
                }
            }
            
        });

        // zip the jars
        for (File jarFile : jarFiles) {
            FileInputStream jarFileStream = new FileInputStream(jarFile);
            ZipEntry zipEntry = new ZipEntry(jarFile.getName());
            zipStream.putNextEntry(zipEntry);
            zipStream.write(jarFileStream.readAllBytes());
            zipStream.closeEntry();
            jarFileStream.close();
            jarFile.delete();
        }
        zipStream.close();

        // remove temp directory, zip is in another directory
        tempDirJars.delete();

        return zipFile.getName();
    }

    /**
     * Reads given resource file as a string.
     *
     * @param fileName path to the resource file
     * @return the file's contents
     * @throws IOException if read fails for any reason
     */
    static String getResourceFileAsString(String fileName) throws IOException {
        ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
        try (InputStream is = classLoader.getResourceAsStream(fileName)) {
            if (is == null) return null;
            try (InputStreamReader isr = new InputStreamReader(is);
                BufferedReader reader = new BufferedReader(isr)) {
                return reader.lines().collect(Collectors.joining(System.lineSeparator()));
            }
        }
    }

    private class ExecutionResult {
        public int exitCode;
        public String stdout;
        public String stderr;
    }
}